--[[
Name: LibSQL-1.0
Revision: $Rev: 22 $
Author(s): Humbedooh
Description: SQL library for Lua
License: All Sporks Reserved!
]]
-----------------------------------------------------------------------
local rev = tonumber(("$Rev: 22 $"):match("(%d+)")) or 5000
if _G.LibSQL and _G.LibSQL.revision >= rev then return end
_G.LibSQL = { revision = rev}
local LibSQL = _G.LibSQL

local LibSQL_Functions = {'ABS', 'ACOS', 'ADDDATE', 'ASCII', 'ASIN', 'ATAN', 'ATAN2', 'AVG', 'BIN', 'BIT_AND', 'BIT_COUNT', 'BIT_LENGTH', 'BIT_OR', 'CAST', 'CEIL', 'CEILING', 'CHAR_LENGTH', 'CHARACTER_LENGTH', 'CONCAT', 'CONCAT_WS', 'CONV', 'CONVERT', 'COS', 'COT', 'COUNT', 'CURDATE', 'CURRENT_DATE', 'CURRENT_TIME', 'CURRENT_TIMESTAMP', 'CURTIME', 'DATE_ADD', 'DATE_FORMAT', 'DATE_SUB', 'DAYNAME', 'DAYOFMONTH', 'DAYOFWEEK', 'DAYOFYEAR', 'DECODE', 'DEGREES', 'ENCODE', 'ENCRYPT', 'EXP', 'EXPORT_SET', 'EXTRACT', 'FIELD', 'FIND_IN_SET', 'FLOOR', 'FORMAT', 'FOUND_ROWS', 'FROM_DAYS', 'FROM_UNIXTIME', 'GREATEST', 'HEX', 'IFNULL', 'INTERVAL', 'ISNULL', 'LASTInsert_ID', 'LCASE', 'LEAST', 'LEFT', 'LENGTH', 'LN', 'LOCATE', 'LOG', 'LOG2', 'LOG10', 'LOWER', 'LPAD', 'LTRIM', 'MAKE_SET', 'MAX', 'MD5', 'MID', 'MIN', 'MOD', 'MONTHNAME', 'NOW', 'NULLIF', 'OCT', 'OCTET_LENGTH', 'ORD', 'PASSWORD', 'PERIOD_ADD', 'PERIOD_DIFF', 'PI', 'POSITION', 'POW', 'POWER', 'QUARTER', 'QUOTE', 'RADIANS', 'RAND', 'RELEASE_LOCK', 'REPEAT', 'REVERSE', 'RIGHT', 'ROUND', 'RPAD', 'RTRIM', 'SEC_TO_TIME', 'SHA', 'SHA1', 'SIGN', 'SIN', 'SPACE', 'SQRT', 'STD', 'STDDEV', 'STRCMP', 'SUBDATE', 'SUBSTRING', 'SUBSTRING_INDEX', 'SUM', 'SYSDATE', 'TAN', 'TIME_FORMAT', 'TIME_TO_SEC', 'TO_DAYS', 'TRIM', 'UCASE', 'UNIX_TIMESTAMP', 'UPPER', 'WEEK', 'WEEKDAY', 'YEARWEEK'}
local LibSQL_Field_Flags = {'AUTO_INCREMENT','BINARY','DEFAULT','NATIONAL','PRECISION','UNSIGNED','VARYING','ZEROFILL', 'NOTNIL', 'UNIQUE', ''}
local LibSQL_Reserved = {'ACTION','ADD','AFTER','AGAINST','AGGREGATE','ALL','ALTER','ANALYSE','ANALYZE','AND','AS','ASC','AUTOCOMMIT','AVG_ROW_LENGTH','BACKUP','BEGIN','BETWEEN','BOTH','BY','CASCADE','CASE','CHANGE','CHANGED','CHECK','CHECKSUM','CLIENT','COLUMN','COLUMNS','COMMENT','COMMIT','COMMITTED','COMPRESSED','CONCURRENT','CONSTRAINT','CREATE','CROSS','DATA','DATABASE','DATABASES','DAY','DAY_HOUR','DAY_MINUTE','DAY_SECOND','DELAYED','DELAY_KEY_WRITE','DELETE','DESC','DESCRIBE','DISTINCT','DISTINCTROW','DO','DROP','DUMPFILE','DYNAMIC','ELSE','ENCLOSED','END','ESCAPE','ESCAPED','EXECUTE','EXISTS','EXPLAIN','EXTENDED','FAST','FIELDS','FILE','FIRST','FIXED','FLUSH','FOR','FOREIGN','FROM','FULL','FULLTEXT','FUNCTION','GEMINI','GEMINI_SPIN_RETRIES','GLOBAL','GRANT','GRANTS','GROUP','HAVING','HEAP','HIGH_PRIORITY','HOSTS','HOUR','HOUR_MINUTE','HOUR_SECOND','IDENTIFIED','IF','IGNORE','IN','INDEX','INDEXES','INFILE','INNER','INSERT','INSERT_ID','INSERT_METHOD','INTERVAL','INTO','IS','ISOLATION','JOIN','KEY','KEYS','KILL','LASTInsert_ID','LEADING','LEFT','LEVEL','LIKE','LIMIT','LINES','LOAD','LOCAL','LOCK','LOCKS','LOGS','LOW_PRIORITY','MASTER','MASTER_CONNECT_RETRY','MASTER_HOST','MASTER_LOG_FILE','MASTER_LOG_POS','MASTER_PASSWORD','MASTER_PORT','MASTER_USER','MATCH','MAX_CONNECTIONS_PER_HOUR','MAX_QUERIES_PER_HOUR','MAX_ROWS','MAX_UPDATES_PER_HOUR','MEDIUM','MERGE','MIN_ROWS','MINUTE','MINUTE_SECOND','MODE','MODIFY','MONTH','MRG_MYISAM','MYISAM','NATURAL','NO','NOT','NULL','ON','OPEN','OPTIMIZE','OPTION','OPTIONALLY','OR','ORDER','OUTER','OUTFILE','PACK_KEYS','PARTIAL','PASSWORD','PRIMARY','PRIVILEGES','PROCEDURE','PROCESS','PROCESSLIST','PURGE','QUICK','RAID0','RAID_CHUNKS','RAID_CHUNKSIZE','RAID_TYPE','READ','REFERENCES','REGEXP','RELOAD','RENAME','REPAIR','REPEATABLE','REPLACE','REPLICATION','RESET','RESTORE','RESTRICT','RETURNS','REVOKE','RIGHT','RLIKE','ROLLBACK','ROW','ROW_FORMAT','ROWS','SECOND','SELECT','SERIALIZABLE','SESSION','SHARE','SHOW','SHUTDOWN','SLAVE','SONAME','SQL_AUTO_IS_NULL','SQL_BIG_RESULT','SQL_BIG_SELECTS','SQL_BIG_TABLES','SQL_BUFFER_RESULT','SQL_LOG_BIN','SQL_LOG_OFF','SQL_LOG_UPDATE','SQL_LOW_PRIORITY_UPDATES','SQL_MAX_JOIN_SIZE','SQL_QUOTE_SHOW_CREATE','SQL_SAFE_UPDATES','SQL_SELECT_LIMIT','SQL_SLAVE_SKIP_COUNTER','SQL_SMALL_RESULT','SQL_WARNINGS','START','STARTING','STATUS','STOP','STRAIGHT_JOIN','STRING','STRIPED','SUPER','TABLE','TABLES','TEMPORARY','TERMINATED','THEN','TO','TRAILING','TRANSACTION','TRUNCATE','TYPE','UNCOMMITTED','UNION','UNIQUE','UNLOCK','UPDATE','USAGE','USE','USING','VALUES','VARIABLES','WHEN','WHERE','WITH','WORK','WRITE','YEAR_MONTH'}
local LibSQL_Field_Types = {'INTEGER', 'NUMBER', 'STRING', 'BOOLEAN', 'ENUM', 'DATE', 'TIMESTAMP', 'BLOB'}
LibSQL_Ascending = 1;
LibSQL_Descending = 2;
LibSQL_AscendingWithCase = 3;
LibSQL_DescendingWithCase = 4;
local  LibSQL_SearchCache = setmetatable({}, {__mode = "v"})
local LibSQL_Statement_Prototypes = {

}

-- [[ UPVALUES ]]--
local _G = _G
local assert = _G.assert
local time = assert(_G.time or _G.os.time)
local upper = assert(_G.string.upper)
local lower = assert(_G.string.lower)
local match = assert(_G.string.match)
local gmatch = assert(_G.string.gmatch)
local len = assert(_G.string.len)
local gsub = assert(_G.string.gsub)
local sub = assert(_G.string.sub)
local find = assert(_G.string.find)
local format = assert(_G.string.format)
local tonumber = assert(_G.tonumber)
local tostring = assert(_G.tostring)
local format = assert(_G.string.format)
local getmetatable = assert(_G.getmetatable)
local setmetatable = assert(_G.setmetatable)
local pairs = assert(_G.pairs)
local next = assert(_G.next)
local type = assert(_G.type)
local rawget = assert(_G.rawget)
local error = assert(_G.error)
local select = assert(_G.select)
local tsort = assert(_G.table.sort)
local tinsert = assert(_G.table.insert)
local tremove = assert(_G.table.remove)
local floor = assert(_G.math.floor)
local sort = assert(_G.table.sort)
local wipe = assert(_G.table.wipe or function(tbl) for n=1, #tbl do tremove(tbl) end end)
local pcall = assert(_G.pcall)
local GetTime = assert(GetTime or os.clock)


--[[ TOOL SET ]]--
local function LibSQL_Validate_Field_Name(name)
    if type(name) ~= "string" or match(name, "[^a-zA-Z0-9_]") then
        return false
    else
        return true
    end
end

local function LibSQL_Set_Weak_Table(tbl)
    setmetatable(tbl, {__mode = "v"})
    return tbl
end

local function LibSQL_Field_Exists(tbl, field)
    local t,x = tbl._fields, lower(field)
    for n = 1, #t do
        if t[n].id == x then
            return t[n]
        end
    end
    return nil
end

local function LibSQL_Field_Valid_Type(field)
    local field = upper(field)
    for n = 1, #LibSQL_Field_Types do
        if LibSQL_Field_Types[n] == field then
            return true
        end
    end
    return false
end

local function LibSQL_Field_Valid_Flag(field)
    local field = upper(field)
    for n = 1, #LibSQL_Field_Flags do
        if LibSQL_Field_Flags[n] == field then
            return true
        end
    end
    return false
end

local function LibSQL_Query_Valid_Function(field)
    local field = upper(field)
    for n = 1, #LibSQL_Functions do
        if LibSQL_Functions[n] == field then
            return true
        end
    end
    return false
end

local function LibSQL_Conform_Data(value, _type)
    if      _type == "NUMBER"    then return tonumber(value) or 0
    elseif  _type == "INTEGER"   then return floor(tonumber(value) or 0)
    elseif  _type == "BOOLEAN"   then return (value == true or lower(value or "") == "true" and true) or false
    elseif  _type == "STRING"    then return (type(value) == "string" and value) or tostring(value)
    end
    return value
end

local function vcopy(tbl)
    if type(tbl) == "table" then
        local ret = {};
        for n = 1, #tbl do
            ret[n] = tbl[n]
        end
    return ret
    else
        return {}
    end
end

local function LibSQL_Resource_Select_Field(tbl, field, as)
    if lower(field) == "from" then return false end
    local field = LibSQL_Field_Exists(tbl, field)
    if field then
        if not as then return field
        else return {pos=field.pos, type=field.type, id=id, xid = as, name=as, flags=field.flags}
        end
    end
end

local function LibSQL_Resource_Select_Function(tbl, a,b,c,d)

end

local function LibSQL_Resource_Update_String(tbl, field, value)
    local field = LibSQL_Field_Exists(tbl, field)
    if field then
        local del = sub(value,1,1)
        if del == [[']] or del == [["]] then
            value = sub(value, 2, len(value)-1)
        end
        value = LibSQL_Conform_Data(value, field.type)
        return unpack({field.id, value})
    end
end

LibSQL_Statement_Prototypes.SELECT = {
    ["^(%a+)%s*as%s*(%a+)[%s,]+"] = LibSQL_Resource_Select_Field,               -- SELECT foo AS bar...
    ["^(%a+)[%s,]+"] = LibSQL_Resource_Select_Field,                            -- SELECT foo, bar...
    ["^(%a+)%s*(%b())%s*as%s*(%a+)[%s,]+"] = LibSQL_Resource_Select_Function,   -- SELECT Foo(bar) AS baz...
    ["^(%a+)%s*(%b())[%s,]+"] = LibSQL_Resource_Select_Function,                -- SELECT Foo(bar)...
}

LibSQL_Statement_Prototypes.UPDATE = {
    ["^(%a+)%s*=%s*([0-9.%-]+)[%s,]+"] = LibSQL_Resource_Update_String,           -- SET foo = 123...
    ["^(%a+)%s*=%s*(%a+)[%s,]+"] = LibSQL_Resource_Update_Field,                -- SET foo = bar...
    ["^(%a+)%s*=%s*(%b\"\")[%s,]+"] = LibSQL_Resource_Update_String,            -- SET foo = "bar"...
    ["^(%a+)%s*=%s*(%b'')[%s,]+"] = LibSQL_Resource_Update_String,              -- SET foo = 'bar'...
    ["^(%a+)%s*=%s*(%a+)(%b())[%s,]+"] = LibSQL_Resource_Select_Function,          -- SET foo = bar(...)...
}

LibSQL_Statement_Prototypes.INSERT = {
    ["^([0-9.%-]+)[%s,]+"] = false,                               -- 123...
    ["^(%b\"\")[%s,]+"] = false,                                  -- "bar"...
    ["^(%b'')[%s,]+"] = false,                                    -- 'bar'...
    ["^(%a+)(%b())[%s,]+"] = LibSQL_Resource_Select_Function,     -- bar(...)...
}


--[[ PARSERS ]]--
local function LibSQL_Parse_Alter(statement)
    local ret = {}
    local cmd = match(statement, "^%s*([%w]+)")
    if cmd then
        ret.cmd = upper(cmd)
        local a,b = find(statement, "(%b())")
        if a and b then statement = sub(statement, a+1,b-1)
        else statement = gsub(statement, "^%s*[%w]+%s*", "", 1)
        end
        for el in gmatch(statement, "([^,]+)%s*,?%s*") do
            ret[#ret+1] = {match(el, "(%w+)%s*(%w*)%s*(.*)")}
        end
    end
    return ret
end


local function LibSQL_ParseFind(tbl, statement)
    if LibSQL_SearchCache[statement] then
        return LibSQL_SearchCache[statement]
    end
    local ret = {vars={},pos={}}
    for var in statement:gmatch("(%w+)") do
        var = lower(var)
        if ( var == "and" or var == "or" or var == "not" or var == "nil" or var == "true" or var == "false" ) then

        elseif LibSQL_Field_Exists(tbl, var) then
        else
        end
    end

    for n = 1, #tbl._fields do
        local field = tbl._fields[n]
        if type(field.pos) == "number" then
            ret.vars[field.pos] = field.id
        else
            tinsert(ret.vars, field.id)
        end
    end
    local f = "local " .. table.concat(ret.vars, ", ") .. " = unpack(_G.__LIBSQL_BUFFER) return (" .. statement .. ") "
    --print(f)
    ret.cmd, err = loadstring(f)
    --if err then error(err) end
    LibSQL_SearchCache[statement] = ret
    return ret
end



--[[ RESOURCE OBJECTs ]]--
LibSQL_Resource = {}
function LibSQL_Slow_Unpack(fields, entry)
    local list = {}
    for key, val in pairs(fields) do
        tinsert(list, entry[val.pos] or entry[val.id])
    end
    return list
end

local function LibSQL_Resource_Next(res)
    res._pointer = (res._pointer or 0)+1
    local entry = res._entries[res._index[res._pointer]]
    if not entry then return nil end
    if #entry ~= 0 then
        return unpack(entry)
    else
        return unpack(LibSQL_Slow_Unpack(res._fields,entry))
    end
end

local function LibSQL_Resource_Next_Arrayed(res)
    res._pointer = (res._pointer or 0)+1
    local entry = res._entries[res._index[res._pointer]]
    if not entry then return nil end
    return entry or {}
end

local function LibSQL_Resource_Next_Objectified(res)
    res._pointer = (res._pointer or 0)+1
    local entry = res._entries[res._index[res._pointer]]
    if not entry then return nil end
    local obj = {}
    local f = res._fields
    for n = 1, #f do
        local x = f[n]
        obj[x.xid or x.id] = entry[x.pos] or entry[x.id]
    end
    return obj
end



function LibSQL_Resource:Insert(...)
    local tbl, fields = self, self._fields
    local num = select('#', ...)
    if num == 1 and type(select(1, ...)) == "string" then
        local list, hash = match(select(1, ...), "(.*)VALUES%s*%(?(.-)%)?%s*$")
        if list then
            local keys, values = {}, {}
            for key in gmatch(list, "(%a+)") do
                tinsert(keys, key)
            end
            if #keys == 0 then
                for k, v in pairs(fields) do tinsert(keys, v.id) end
            end
            local protos = LibSQL_Statement_Prototypes.INSERT
            local pos = 1
            local ppos = 1
            local i = 0
            local field
            hash = hash .. ", "
            while pos do
                ppos = pos
                i = i+1
                for key, func in pairs(protos) do
                    local x,y,a,b,c,d = find(hash, key, pos)
                    if y then
                        pos = y+1
                        if func then value = func(self,a,b,c,d)
                        else value = a
                        end
                        local del = sub(value,1,1)
                        if del == [[']] or del == [["]] then
                            value = sub(value, 2, len(value)-1)
                        end
                        values[keys[i]] = value
                        break
                    end
                end
                if ppos == pos then break end
            end
            self:Insert(values)
        end
    else
        for n = 1, num do
            local hash = select(n, ...)
            if type(hash) == "table" then
                local entry
                if hash[1] then
                    for key, field in pairs(fields) do
                        local value = hash[key]
                        if value ~= nil then
                            hash[field.pos] = LibSQL_Conform_Data(value, field.type)
                        else
                            if field.flags == "AUTO_INCREMENT" then
                                tbl._autoid = tbl._autoid + 1
                                hash[field.pos] = tbl._autoid
                            else
                                hash[field.pos] = nil
                            end
                        end
                    end
                    entry = hash
                else
                    entry = {}
                    local tmp = {}
                    for field, value in pairs(hash) do
                        tmp[lower(field)] = value
                    end
                    for key, field in pairs(fields) do
                        local value = tmp[field.id]
                        if value ~= nil then
                            entry[field.pos] = LibSQL_Conform_Data(value, field.type)
                        else
                            if field.flags == "AUTO_INCREMENT" then
                                tbl._autoid = tbl._autoid + 1
                                entry[field.pos] = tbl._autoid
                            else
                                entry[field.pos] = nil
                            end
                        end
                    end
                end
                tinsert(tbl._entries, entry)
                tinsert(tbl._index, #tbl._entries)
            end
        end
    end
end

function LibSQL_Resource:Update(hash)
    local tbl, fields, entries = self, self._fields, self._entries
    if type(hash) == "string" then
        hash = hash .. ", "
        local index = {}
        local protos = LibSQL_Statement_Prototypes.UPDATE
        local pos = 1
        local ppos = 1
        local field
        while pos do
            ppos = pos
            for key, func in pairs(protos) do
                local x,y,a,b,c,d = find(hash, key, pos)
                if y then
                    pos = y+1
                    field, value = func(self,a,b,c,d)
                    if field then index[field] = value end
                    break
                end
            end
            if ppos == pos then break end
        end
        hash = index
    end
    if type(hash) == "table" then
        for key, value in pairs(hash) do
            local field = LibSQL_Field_Exists(tbl, key or "")
            if field then
                for _, entry in pairs(tbl._index) do
                    entries[entry][field.pos] = LibSQL_Conform_Data(value, field.type)
                end
            end
        end
    end
end

function LibSQL_Resource:Remove(a,b)
    local tbl = self
    sort(tbl._index)
    for n = #tbl._index, 1, -1 do
        local x = tbl._index[n]
        tremove(tbl._entries,x)
    end
    tbl._index = {}
end

function LibSQL_Resource:Find(statement)
    local res = self;
    local ret = LibSQL_ParseFind(res, statement)
    local nres = LibSQL_Resource:New(res);
    local func = ret.cmd
    local ent,ind,nind = res._entries, res._index, nres._index
    local _G = _G
    for _, n in pairs(ind) do
        if #ent[n] ~= 0 then
            _G.__LIBSQL_BUFFER = ent[n]
        else
            _G.__LIBSQL_BUFFER = LibSQL_Slow_Unpack(res._fields,ent[n])
        end
        local mkay, positive = pcall(func)
        if mkay and positive then
            nind[#nind+1] = n
        else
--            print(positive)
        end
    end
    return nres
end

function LibSQL_Resource:Alter(statement)
    local args = LibSQL_Parse_Alter(statement)
    local tbl = self
    local cmd = args.cmd;
    if 		cmd == "ADD"    then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' already exists!", args[n][1])) return
            else
                local t,f = upper(args[n][2]), upper(args[n][3])
                if not LibSQL_Field_Valid_Type(t) then error(format("LibSQL: Invalid field type '%s'!", t)) return end
                if not LibSQL_Field_Valid_Flag(f) then error(format("LibSQL: Invalid field flag '%s'!", f)) return end
            end
        end
        for n = 1, #args do
            local n,t,f = args[n][1], upper(args[n][2]), upper(args[n][3])
            tinsert(tbl._fields, {pos=(#tbl._fields)+1, id=lower(n), name=n, flags=f,type=t})
        end
    elseif	cmd == "DROP"   then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if not LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' does not exist!", args[n][1])) return end
        end
        local t = tbl._fields
        for n = 1, #args do
            local x = lower(args[n][2]) for n = 1, #t do if t[n].id == x then tremove(t, n) break end end
        end
        for n = 1, #t do
            t[n].pos = n
        end
    elseif cmd == "RENAME"	then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if not LibSQL_Validate_Field_Name(args[n][2]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][2])) return end
            if not LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' does not exist!", args[n][1])) return end
            if not LibSQL_Field_Exists(tbl, args[n][2]) then error(format("LibSQL: Field '%s' already exist!", args[n][2])) return end
        end
        for n = 1, #args do
            local x = lower(args[n][2]) for n = 1, #t do if t[n].id == x then t[n].id = x t[n].name = args[n][2] break end end
        end
    elseif cmd == "MODIFY"	then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if not LibSQL_Validate_Field_Type(args[n][2]) then error(format("LibSQL: Field type '%s' is invalid!", args[n][2])) return end
            if not LibSQL_Validate_Field_Flag(args[n][3]) then error(format("LibSQL: Field flag '%s' is invalid!", args[n][3])) return end
            if not LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' does not exist!", args[n][1])) return end
        end
        for n = 1, #args do
            local x = lower(args[n][1]) for n = 1, #t do if t[n].id == x then t[n].type = upper(args[n][2]) break end end
        end
    end
    return self
end

function LibSQL_Resource:Limit(a,b)
    local tbl = self;
    local a,b = tonumber(a),tonumber(b)
    local index = {}
    if not b then a,b = 0, a end
    if a and b then
        for n = a+1, a+b do
            tinsert(index, tbl._index[n])
        end
    end
    tbl._index = index
    return self
end


function LibSQL_Resource:Select(statement)
    if match(statement, "^%s**%s*") then return self end
    statement = statement .. ", "
    local index = {}
    local protos = LibSQL_Statement_Prototypes.SELECT
    local pos = 1
    local ppos = 1
    local field
    while pos do
        ppos = pos
        for key, func in pairs(protos) do
            local x,y,a,b,c,d = find(statement, key, pos)
            if y then
                pos = y+1
                field = func(self,a,b,c,d)
                if field then tinsert(index, field) end
                break
            end
        end
        if ppos == pos then break end
    end
    self._fields = index
    return self
end


local function SortFunc() return false end

-- fx. Sort({name = LibSQL_Ascending, id = LibSQL_Descending})
function LibSQL_Resource:Sort(hash)
    if type(hash) == "string" then
        local nhash = {}
        for field, dir in gmatch(hash, "(%a+) (%a+),?") do
            nhash[field] = (lower(dir) == "desc" and LibSQL_Descending) or LibSQL_Ascending
        end
        hash = nhash
    end
    if type(hash) == "table" then
        local func = SortFunc
        local tbl = self._entries
        local rhash = {}
        for field, dir in pairs(hash) do
            tinsert(rhash, {field, dir})
        end
        for n = #rhash, 1, -1 do
            local field, dir = unpack(rhash[n])
            local obj = LibSQL_Field_Exists(self, field)
            local l = (obj.type == "STRING" and true) or false
            local r = true
            if      ( dir == LibSQL_AscendingWithCase ) then r = true  l = false
            elseif  ( dir == LibSQL_Descending )        then r = false
            elseif  ( dir == LibSQL_AscendingWithCase ) then r = false l = false
            end

            local c = obj.pos
            local myFunc = func;
            if l then
                func = function(a,b)
                        if a == nil or b == nil then return nil; end
                        local x,y = lower(tbl[a][c]) or 0, lower(tbl[b][c]) or 0
                        if x < y then return r
                        elseif x == y then
                            return myFunc(a,b)
                        end
                        return not r
                    end
            else
                func = function(a,b)
                        if a == nil or b == nil then return nil end
                        local x,y = tbl[a][c] or 0, tbl[b][c] or 0
                        if x < y then return r
                        elseif x == y then
                            return myFunc(a,b)
                        end
                        return not r
                    end
            end
        end
        table.sort(self._index, func);
        return self
    end
end

function LibSQL_Resource:Flush() self._pointer = nil self._entries = {} self._fields = {} end
function LibSQL_Resource:Rewind() self._pointer = nil end
function LibSQL_Resource:Seek(num) self._pointer = tonumber(num) return self._pointer end
function LibSQL_Resource:Rows() return #self._index end
function LibSQL_Resource:Fetch() return LibSQL_Resource_Next, self, nil end
function LibSQL_Resource:FetchObject() return LibSQL_Resource_Next_Objectified, self, nil end
function LibSQL_Resource:FetchArray() return LibSQL_Resource_Next_Arrayed, self, nil end

function LibSQL_Resource:New(hash, copy)
    local res = {}
    res.Fetch =             LibSQL_Resource.Fetch
    res.FetchObject =       LibSQL_Resource.FetchObject
    res.FetchArray =        LibSQL_Resource.FetchArray
    res.Rewind =            LibSQL_Resource.Rewind
    res.Seek =              LibSQL_Resource.Seek
    res.Rows =              LibSQL_Resource.Rows
    res.Flush =				LibSQL_Resource.Flush
    res.Alter =             LibSQL_Resource.Alter
    res.Find =              LibSQL_Resource.Find
    res.Remove =            LibSQL_Resource.Remove
    res.Find =              LibSQL_Resource.Find
    res.Insert =            LibSQL_Resource.Insert
    res.Sort =              LibSQL_Resource.Sort
    res.Limit =             LibSQL_Resource.Limit
    res.Select =            LibSQL_Resource.Select
    res.Update =            LibSQL_Resource.Update

    -- "Protected" functions
    if type(hash) ~= "table" then hash = {} end
    res._fields =   hash._fields or LibSQL_Set_Weak_Table({})
    res._entries =  hash._entries or LibSQL_Set_Weak_Table({})
    res._parent = hash
    res._index = {}
    res._autoid = hash._autoid or 0
    if copy then
        for n = 1, #res._entries do
            res._index[n] = n
        end
    end
    res._pointer = nil
    return res
end

--[[ DATABASE OBJECTS ]]--

-- example:
-- CREATE TABLE myTable (name STRING UNIQUE, level NUMBER, id NUMBER AUTO_INCREMENT);
-- Translates to;
-- db:Create("myTable"):Alter("ADD (name STRING UNIQUE, level NUMBER, id NUMBER AUTO_INCREMENT)")
function LibSQL:Create(tbl)
    if LibSQL_Validate_Field_Name(tbl) then
        local db = self
        local tblint = lower(tbl)
        if db._tables[tblint] then error(format(tbl,"LibSQL: Table %s already exists!")) return end
        db._tables[tblint] = {created = time(), name = tbl, insert_id = 0, _fields = {}, _entries = {}, _autoid = 0}
        return LibSQL_Resource:New(db._tables[tblint], true)
    end
end

function LibSQL:Drop(tbl)
    if LibSQL_Validate_Field_Name(tbl) then
        local db = self
        local tblint = lower(tbl)
        if not db._tables[tblint] then error(format("LibSQL: Table '%s' doesn't exist!", tbl)) return end
        db._tables[tblint] = nil;
        return true;
    end
end



function LibSQL:Get(tbl)
    local db = self
    if LibSQL_Validate_Field_Name(tbl) then
        local tblhash = db._tables[lower(tbl)]
        if tblhash then return LibSQL_Resource:New(tblhash, true)
        else error(format("LibSQL: Unknown table '%s'!", tbl)) return
        end
    end
    error(format("LibSQL: Invalid table name '%s'!", tbl or "??"))
end

function LibSQL:_Parse(arg)
    local db = self
    if type(arg) ~= "string" or type(db) ~= "table" then return end
    local cmd, statement = match(arg,"^(%a+)%s*(.*)")
    if cmd and statement then
        cmd = cmd:upper()
        if      cmd == "CREATE"     then
            local tbl, statement = match(statement, "(%a+)%s+(.*)")
            return db:Create(tbl):Alter("ADD " .. statement)
        elseif  cmd == "SELECT"     then
            local x,y, selection, tbl, clause = find(statement, "^(.*)FROM%s*(%a+)%s*WHERE(.-)$")
            if x and y then
                local x,y,limita, limitb = find(clause, "LIMIT ([0-9]+),?([0-9]*)")
                if x then clause = sub(clause, 1, x-2) end
                local x,y,sorting = find(clause, "SORT BY (.*)$")
                if x then clause = sub(clause, 1, x-2) end
                local tbl = db:Get(tbl):Find(clause)
                if sorting then tbl:Sort(sorting) end
                if limita then tbl:Limit(limita, limitb) end
                return tbl:Select(selection)
            end
            -- Done
        elseif  cmd == "INSERT"     then
            local x,y, tbl, hash = find(statement, "^%s*INTO%s*(%a+)%s*(.*)$")
            return db:Get(tbl):Insert(hash)
            -- Done
        elseif  cmd == "UPDATE"     then
            local x,y, tbl, set, clause = find(statement, "^%s*(%a+)%s*SET%s*(.*)%s*WHERE(.*)$")
            if x and y then
                local x,y,limita, limitb = find(clause, "LIMIT ([0-9]+),?([0-9]*)")
                if x then clause = sub(clause, 1, x-2) end
                local x,y,sorting = find(clause, "SORT BY (.*)$")
                if x then clause = sub(clause, 1, x-2) end
                local tbl = db:Get(tbl):Find(clause)
                if sorting then tbl:Sort(sorting) end
                if limita then tbl:Limit(limita, limitb) end
                tbl:Update(set)
            end -- Done
        elseif  cmd == "REMOVE"     then
            local x,y, selection, tbl, clause = find(statement, "^(.*)FROM%s*(%a+)%s*WHERE(.-)$")
            if x and y then
                local x,y,limita, limitb = find(clause, "LIMIT ([0-9]+),?([0-9]*)")
                if x then clause = sub(clause, 1, x-2) end
                local x,y,sorting = find(clause, "SORT BY (.*)$")
                if x then clause = sub(clause, 1, x-2) end
                local tbl = db:Get(tbl):Find(clause)
                if sorting then tbl:Sort(sorting) end
                if limita then tbl:Limit(limita, limitb) end
                tbl:Remove()
            end
            -- Done
        elseif  cmd == "ALTER"      then
            local tbl, statement = match(statement, "(%a+)%s+(.*)")
            return db:Get(tbl):Alter(statement)
            -- Done
        elseif  cmd == "TRUNCATE"   then return db:Get(statement):Remove() -- Done
        --elseif  cmd == "FOLD"       then return db:Get(tbl):Fold(statement)
        elseif  cmd == "DROP"       then return db:Drop(statement) -- Done
        end
    end
end

function LibSQL:Execute(...)
    local n = select("#",...)
    if n == 1 then return self:_Parse(...)
    else    return self:_ParseInject(...)
    end
end


function LibSQL:Acquire(name, data, args)
    if type(name) == "string" and type(data) == "table" and type(args) == "table" then
        local tbl = self:Create(name)
        tbl._parent._entries = data
        for key, arg in pairs(args) do
            local name = match(arg, "AS%s+(%a+)") or key
            pos = tonumber(match(arg, "AT%s+(%d+)")) or name
            local t = match(arg, "^%s*(%d+)") or "STRING"
            id = name
            if type(name) == "string" then id = lower(id) end
            tinsert(tbl._fields, {id = id, name = name, pos = pos, type = t})
        end
    end
end

function LibSQL:Open(db)
    if type(db) ~= "table" or not db._tables then error("Usage: LibSQL:Open(database)") return end
    db.Execute = 	LibSQL.Execute
    db.Get = 		LibSQL.Get
    db.Create =	LibSQL.Create
    db._Parse =     LibSQL._Parse
    db.Drop =      LibSQL.Drop
    db.Acquire = LibSQL.Acquire
    return db
end

function LibSQL:New() return LibSQL:Open({_tables = {}}) end


local debug = false
if debug then
    local db = LibSQL:New()
    db:Execute("CREATE myTable (name STRING UNIQUE, level INTEGER, id INTEGER AUTO_INCREMENT)")
    local tbl = db:Get("myTable")
    local x = GetTime()
    for n = 1, 5 do
    tbl:Insert(
        {name="Abe", level=3},
        {name="Benny", level=2.4, bleh = 3},
        {"Benny", 4.5},
        {"Charles", 5},
        {"Abe", 4}
    )
    end
    --tbl:Update({level = 6})
    --tbl:Update("level = 7, name = 'moo', id = count()")
    db:Execute("UPDATE myTable SET name = 'moo', level = 46 WHERE level < 4 LIMIT 2")

    db:Execute("REMOVE FROM myTable WHERE level > 5")
    db:Execute("INSERT INTO myTable VALUES ('smurf', 100)")
    local res = db:Execute("SELECT * FROM myTable WHERE level > 1 SORT BY name DESC LIMIT 5")
    for i = 1, 100000 do
        res:Sort( { name = LibSQL_Ascending, level = LibSQL_Descending})
    end

    print(GetTime()-x)
    print("got", res:Rows(), "results!\n")
    for a,b,c,d in res:FetchObject() do
        print(a.name,a.level,a.id,d)
    end

    db:Execute("TRUNCATE myTable")
    db:Execute("DROP myTable")
end

